-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_blake_package.ALL;
use work.sha3_pkg.all;

entity blake_datapath_ppl is	 	
	generic (	  	   
		FF : integer := 4;	
		PPL : integer := 4;
		BS : integer := 512;
		W : integer := 64;
		IW : integer := 32;
		HS : integer := 256
		);
	port (
		-- external
		rst 				: in std_logic;
		clk					: in std_logic;
		din  				: in std_logic_vector(ppl*w-1 downto 0);	
		dout 				: out std_logic_vector(w-1 downto 0);	
		--fsm 1		
		ein 				: in std_logic_vector(ppl-1 downto 0);
		ec, lc, lm	 		: in std_logic_vector(ppl-1 downto 0);
		dt, eth, etl 		: in std_logic_vector(ppl-1 downto 0);		
		sel_t				: in std_logic_vector(PPL-1 downto 0);
		zc0, zc1 			: out std_logic_vector(ppl-1 downto 0);
		
		-- pad
		last_word 			: in std_logic_vector(PPL-1 downto 0);
		en_lb				: in std_logic_vector(PPL-1 downto 0);
		clr_lb				: in std_logic_vector(PPL-1 downto 0);
		sel_pad				: in std_logic_vector(PPL*(HS/256)-1 downto 0);
		spos 				: in std_logic_vector(PPL*2-1 downto 0);	 
		comp_rem_e0			: out std_logic_vector(PPL-1 downto 0);
		comp_lb_e0			: out std_logic_vector(PPL-1 downto 0);
		comp_rem_mt440 		: out std_logic_vector(PPL-1 downto 0);		
		
		--fsm 2	
		load_h 				: in std_logic;
		eh 					: in std_logic_vector(ppl-1 downto 0);
		smsg 				: in std_logic_vector(log2(ppl)-1 downto 0);
		round 				: in std_logic_vector(3+FF-1 downto 0);				
		sf 					: in std_logic;
		sperm 				: in std_logic;
		
		-- Output				
		shift_dout, load_id, load_hash 	: in std_logic_vector(ppl-1 downto 0);
		sout 							: in std_logic_vector(log2(ppl)-1 downto 0);
		sid 							: in std_logic
	);				  
end blake_datapath_ppl;

architecture struct of blake_datapath_ppl is   
	
	-- CONSTANTS
	constant mw 			: integer := BS;		 
	constant bseg			: integer := BS/w;		  
	constant log2b			: integer := log2( BS );
	constant log2bzeros		: std_logic_vector(log2b-1 downto 0) := (others => '0');	
	constant bzeros			: std_logic_vector(BS-1 downto 0) := (others => '0');
	constant bhalfzeros		: std_logic_vector(BS/2-1 downto 0) := (others => '0');
	constant bmin	: integer := BS - 2 - iw*2;
	constant iv : std_logic_vector(BS/2-1 downto 0) := get_iv( HS, IW );	
	type bot_permute_type is array ( 0 to 15 ) of integer;
	constant bot_permute : bot_permute_type := ( 0,1,2,3,5,6,7,4,10,11,8,9,15,12,13,14 );	
	constant blake_constant : std_logic_vector(BS-1 downto 0) := get_cons( HS, BS, iw );
	-------------
	
	-- //////////////////////
	-- FUNCTIONS 
	type std_logic_matrix is array (natural range <>) of std_logic_vector(iw - 1 downto 0) ;
	function wordmatrix2blk  	(x : std_logic_matrix) return std_logic_vector is
		variable retval : std_logic_vector(BS-1 downto 0);
	begin
		for i in 0 to 15 loop
			retval(iw*(i+1) - 1 downto iw*i) := x(15-i);
		end loop;
		return retval;
	end wordmatrix2blk;
	function blk2wordmatrix  	(x : std_logic_vector; blksize : integer ) return std_logic_matrix is
		variable retval : std_logic_matrix(0 to blksize-1);
	begin
		for i in 0 to blksize-1 loop
			retval(blksize-1-i) := x(iw*(i+1) - 1 downto iw*i);
		end loop;
		return retval;
	end blk2wordmatrix;	
	-------------
	
	-- //////////////////////
	-- type declaration
	type c_array_type is array ( 0 to ppl-1) of std_logic_vector(15 downto 0);	
	type block_array_type is array (0 to ppl-1) of std_logic_vector(BS-1 downto 0);
	type half_block_array_type is array (0 to ppl-1) of std_logic_vector(BS/2-1 downto 0);	
	type t_top_type is array ( 0 to ppl-1 ) of std_logic_vector(2*iw-log2b-1 downto 0);
	type t_bot_type is array ( 0 to ppl-1 ) of std_logic_vector(log2b-1 downto 0);
	type t_type is array ( 0 to ppl-1 ) of std_logic_vector(2*iw-1 downto 0);
	-------------

	signal v1, v2, v2_perm, v2_revert, v2_regged, v3 : std_logic_matrix( 0 to 15 ); 
	signal cp : std_logic_matrix(0 to 16/FF - 1);	 
	
	-- FSM 1 (PROTOCOL / LOADING )
	signal c : c_array_type;
	signal t_top : t_top_type;
	signal t_bot : t_bot_type;
	signal t, t_original : t_type;
	signal min, m : block_array_type;
	
	-- Padding related signals
	constant zeros : std_logic_vector(BS-1 downto 0) := (others => '0');
	constant BITADD : integer := HS/256-1;			
	
	type rem_bit_type is array (0 to PPL-1) of std_logic_vector(8+BITADD downto 0);
	type ppl_8bit_type is array (0 to PPL-1) of std_logic_vector(7 downto 0);
	type ppl_wbit_type is array (0 to PPL-1) of std_logic_vector(63 downto 0);
	type ppl_sel512_type is array (0 to PPL-1) of std_logic_vector(1 downto 0);
	signal remaining_bits : rem_bit_type;
	signal din_padded : ppl_wbit_type;
	signal din_or_length : ppl_wbit_type;
	signal sel_pad_location_lookup, sel_din_lookup : ppl_8bit_type;
	signal sel_din, sel_pad_location : ppl_8bit_type; 
	signal sel_pad_h512 : ppl_sel512_type;
	
	
	-- FSM2 ( ROUND )
	signal t_selected : std_logic_vector(2*iw-1 downto 0);
	signal m_selected : std_logic_vector(BS-1 downto 0);
	signal consout : std_logic_vector(BS/FF-1 downto 0);  			  -- 8 = b size, 4 = b/2 size	
	signal rdprime, rmux,chain_selected : std_logic_vector(BS/2 - 1 downto 0);
	signal rinit, rin, r, rprime, rprime_regged : std_logic_vector(BS-1 downto 0);
	signal chain : half_block_array_type;
	
	-- FSM 3 ( PROTOCOL / OUTPUT )
	-- id, hash		
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
		
	-- debug			  
	-- signal rinit_m, r_m, cp_m : std_logic_matrix(0 to 15);				
begin		
	-- debug
	-- rinit_m <= blk2wordmatrix( rin, 16 );		 
	-- r_m <= blk2wordmatrix( r, 16 );	
	--		

	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION
	fsm1_section : for i in 0 to ppl-1 generate		  
		decounter_gen : decountern generic map ( N => 16, sub => 1 ) 
			port map ( 	clk => clk, rst => '0', load => lc(i), en => ec(i), 
						input => din(log2b+15+(ppl*w-(i+1)*w) downto log2b+(ppl*w-(i+1)*w)), output => c(i) );
		zc0(i) <= '1' when c(i) = 0  else '0'; 	
		zc1(i) <= '1' when c(i) = 1  else '0'; 	
		
		rem1_gen : 
			decountern 
			generic map ( N => 3 + BITADD, sub => 1 ) 
			port map ( clk => clk, rst => clr_lb(i), load => lc(i), en => en_lb(i), 
					  input => din(8 + BITADD + (PPL-i-1)*W downto 6 + (PPL-i-1)*W), 
					  output => remaining_bits(i)(8 + BITADD downto 6));	
			
		rem2_reg : 
			regn generic map ( N => 6, init => "000000" ) 
			port map ( clk => clk, rst => clr_lb(i), en => lc(i), 
					   input => din(5 + (PPL-i-1)*W downto 0 + (PPL-i-1)*W), 
					   output => remaining_bits(i)(5 downto 0));		 	
			
		comp_rem_e0(i) <= '1' when remaining_bits(i) = 0 else '0';
		comp_lb_e0(i) <= '1' when remaining_bits(i)(8+BITADD downto 6) = zeros(2+BITADD downto 0) else '0';	   
		
		sel_pad_location_lookup(i) <= lookup_sel_lvl2_64bit_pad(conv_integer(remaining_bits(i)(5 downto 3)));
		sel_pad_location(i) <= (others => '0') when spos(0 + (PPL-i-1)*2) = '0' else sel_pad_location_lookup(i);
		
		sel_din_lookup(i) <= lookup_sel_lvl1_64bit_pad(conv_integer(remaining_bits(i)(5 downto 3)));	
		sel_din(i) <= (others => '1') when spos(1 + (PPL-i-1)*2) = '1' else sel_din_lookup(i);
	
		byte_pad_gen : 
			entity work.blake_bytepad(struct) 
			generic map ( w => 64 ) 
			port map ( 	din => din((PPL-i)*W-1 downto (PPL-i-1)*W), dout => din_padded(i), 
						sel_pad_location => sel_pad_location(i), sel_din =>  sel_din(i), last_word => last_word(i));
																																		 																														 
		h256_gen : 
			if HS = 256 generate
				din_or_length(i) <= din_padded(i) when sel_pad((PPL-i)*(HS/256)-1 downto (PPL-i-1)*(HS/256)) = "0" else t_original(i)(63 downto 0);				
				comp_rem_mt440(i) <= '1' when remaining_bits(i) > 440 else '0';
			end generate;
		h512_gen : 
			if HS = 512 generate						 
				sel_pad_h512(i) <= sel_pad((PPL-i)*(HS/256)-1 downto (PPL-i-1)*(HS/256));
				with sel_pad_h512(i) select
				din_or_length(i)(63 downto 0) <= t_original(i)(63 downto 0) when "11",
												t_original(i)(127 downto 64) when "10",
												din_padded(i) when others;
				comp_rem_mt440(i) <= '1' when remaining_bits(i) > 888 else '0';
			end generate;	
		 
		-- Len Counter
		th_gen : countern generic map ( N => 2*iw-log2b, step => 1 ) 
			port map ( 	clk => clk, rst => dt(i), load => '0', en => eth(i), 				 
						input => zeros(2*iw-log2b-1 downto 0), output => t_top(i) );
		tl_gen : regn generic map ( N => log2b, init => log2bzeros ) 
			port map ( 	clk => clk, rst => dt(i), en => etl(i), 
						input => remaining_bits(i)(log2b-1 downto 0), output => t_bot(i) );
		t_original(i) <= t_top(i) & t_bot(i);
		t(i) <= t_original(i) when sel_t(i) = '0' else (others => '0');
							 
		shfin_gen : sipo generic map ( N => BS,M =>w) 
			port map (	clk => clk, en => ein(i), 
						input => din_or_length(i), output => min(i) );						
		mreg : regn generic map (N => mw, init => bzeros) 
			port map (clk => clk, rst => '0', en => lm(i), input => min(i), output => m(i) );
	end generate;

	
		
	-- ///////////////////////////////////////////////////////
	-- FSM2 SECTION
	ppl2_sel : if ppl = 2 generate
		t_selected <= t(0) when smsg(0) = '0' else t(1);
		m_selected <= m(0) when smsg(0) = '0' else m(1);
		chain_selected <= chain(0) when smsg(0) = '0' else chain(1);
	end generate;  
	ppl4_sel : if ppl = 4 generate
		with smsg(1 downto 0) select
		t_selected 	<= t(0) when "00", t(1) when "01", t(2) when "10",t(3) when others;	
		with smsg(1 downto 0) select
		m_selected 	<= m(0) when "00", m(1) when "01", m(2) when "10",m(3) when others;	
		with smsg(1 downto 0) select
		chain_selected 	<= chain(0) when "00", chain(1) when "01", chain(2) when "10",chain(3) when others;
	end generate;
	
	rmux <= iv when sf = '1' else rdprime;

	rinit <= rmux & blake_constant(BS-1 downto BS-iw*4)  & 
			(t_selected(iw-1 downto 0) xor blake_constant(BS-1-iw*4 downto BS-iw*5)) &  (t_selected(iw-1 downto 0) xor blake_constant(BS-iw*5-1 downto BS-iw*6)) &
			(t_selected(2*iw-1 downto iw) xor blake_constant(BS-1-iw*6 downto BS-iw*7))  & (t_selected(2*iw-1 downto iw) xor blake_constant(BS-iw*7-1 downto BS-iw*8));
	
	fsm2_section : for i in 0 to ppl-1 generate
		chaining_value_gen : regn generic map ( N => BS/2, init => bhalfzeros ) 
			port map (clk => clk, rst => '0', en => eh(i), input => rmux, output => chain(i) );
	end generate;
	
	rin <= rinit when (load_h = '1') else rprime;
	r_gen : regn generic map ( N => BS, init => bzeros ) 
		port map (clk => clk, rst => '0', en => '1', input => rin, output => r );
	
	-- //////////////
	-- CORE OPERATION	
	g_gfunc4 : if FF = 2 generate
		v1 <= blk2wordmatrix(r, 16);	
		perm4_gen : entity work.permute4xor_ppl(muxbased) 
			generic map (ppl => ppl, h => HS, b => BS, iw => iw) 
			port map ( clk => clk, m => m_selected, round => round, consout => consout );
		cp <= blk2wordmatrix(consout,8);	
		glvl1 : for i in 0 to 3 generate		   
			ppl2 : if ppl = 2 generate
				g0123 : entity work.gfunc_modified_ppl2(struct) 	generic map ( iw => iw, h => HS )
							port map (clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));								
			end generate;
			ppl4 : if ppl = 4 generate 
				g0123 : entity work.gfunc_modified_ppl4(struct) 	generic map ( iw => iw, h => HS )
							port map (clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));								
			end generate;
		end generate;	
		v2_gen : for i in 0 to 15 generate
			v2_perm( bot_permute(i) ) <= v2( i ); 
			v2_revert( i ) <= v2( bot_permute(i) );
		end generate;	
		v3 <= v2_perm when sperm = '0' else v2_revert;		 
	end generate;	
	
	g_gfunc8 : if FF = 1 generate
		perm8_gen : entity work.permute8xor_ppl(muxbased) 
			generic map (ppl => ppl, h => HS, b => BS, iw => iw) 
			port map ( clk => clk,  m => m_selected, round => round, consout => consout );
				
		v1 <= blk2wordmatrix(r, 16);
		cp <= blk2wordmatrix(consout, 16);
		
		v2reg : process( clk ) 
		begin	
			if rising_edge( clk ) then
				v2_regged <= v2;
			end if;
		end process;
			
		ppl4: if ppl = 4 generate
			glvl1 : for i in 0 to 3 generate
				g0123 : entity work.gfunc8x_modified_ppl2(struct) 	generic map ( iw => iw, h => HS )
					port map ( 		clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));
			end generate;
							
			
			
			g4 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
					port map ( 	clk => clk,
								ain => v2_regged(0),bin => v2_regged(5),cin => v2_regged(10),din => v2_regged(15),
								const_0 => cp(8),const_1 => cp(9),
								aout =>  v3(0), bout => v3(5), cout=> v3(10),dout => v3(15));
															 
			g5 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
								ain => v2_regged(1),bin => v2_regged(6),cin => v2_regged(11),din => v2_regged(12),
								 const_0 => cp(10),const_1 => cp(11),
								  aout =>  v3(1), bout => v3(6), cout=> v3(11),dout => v3(12));
		
			g6 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
									ain => v2_regged(2),bin => v2_regged(7),cin => v2_regged(8),din => v2_regged(13),
								  const_0 => cp(12),const_1 => cp(13),
								  aout =>  v3(2), bout => v3(7), cout=> v3(8),dout => v3(13));
		
			g7 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
								ain => v2_regged(3),bin => v2_regged(4),cin => v2_regged(9),din => v2_regged(14),
								  const_0 => cp(14),const_1 => cp(15),
								  aout =>  v3(3), bout => v3(4), cout=> v3(9),dout => v3(14));
		end generate;	
		
		ppl2: if ppl = 2 generate
			glvl1 : for i in 0 to 3 generate
				g0123 : entity work.gfunc_modified(struct) 	generic map ( iw => iw, h => HS )
								port map ( ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									  const_0 => cp(2*i),const_1 => cp(2*i + 1),
									  aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));
			end generate;
							
			
			
			g4 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(0),bin => v2_regged(5),cin => v2_regged(10),din => v2_regged(15),
								  const_0 => cp(8),const_1 => cp(9),
								  aout =>  v3(0), bout => v3(5), cout=> v3(10),dout => v3(15));
															 
			g5 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(1),bin => v2_regged(6),cin => v2_regged(11),din => v2_regged(12),
								 const_0 => cp(10),const_1 => cp(11),
								  aout =>  v3(1), bout => v3(6), cout=> v3(11),dout => v3(12));
		
			g6 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
			 		   port map ( ain => v2_regged(2),bin => v2_regged(7),cin => v2_regged(8),din => v2_regged(13),
								  const_0 => cp(12),const_1 => cp(13),
								  aout =>  v3(2), bout => v3(7), cout=> v3(8),dout => v3(13));
		
			g7 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(3),bin => v2_regged(4),cin => v2_regged(9),din => v2_regged(14),
								  const_0 => cp(14),const_1 => cp(15),
								  aout =>  v3(3), bout => v3(4), cout=> v3(9),dout => v3(14));
		end generate;
		
								  
		rprime <= wordmatrix2blk(v3);
	end generate;
	
	rprime <= wordmatrix2blk(v3);		
	
	rpr : process ( clk ) 
	begin	
		if rising_edge( clk ) then
			rprime_regged <= rprime;
		end if;
	end process;
		
	--finalization				
	rdprime <= chain_selected xor rprime_regged(BS-1 downto BS/2) xor rprime_regged(BS/2-1 downto 0);
	
	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation		 
	
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => rdprime, id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
					
					
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 		
								 
end struct;